@@ -1,5 +1,5 @@
 /* search.c - searching subroutines using dfa, kwset and regex for grep.
-   Copyright 1992, 1998, 2000, 2007, 2009-2016 Free Software Foundation, Inc.
+   Copyright 1992, 1998, 2000, 2007, 2009-2018 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -46,23 +46,29 @@ _GL_INLINE_HEADER_BEGIN
 typedef signed char mb_len_map_t;
 
 /* searchutils.c */
+extern void wordinit (void);
 extern kwset_t kwsinit (bool);
+extern size_t wordchars_size (char const *, char const *) _GL_ATTRIBUTE_PURE;
+extern size_t wordchar_next (char const *, char const *) _GL_ATTRIBUTE_PURE;
+extern size_t wordchar_prev (char const *, char const *, char const *)
+  _GL_ATTRIBUTE_PURE;
 extern ptrdiff_t mb_goback (char const **, char const *, char const *);
-extern wint_t mb_prev_wc (char const *, char const *, char const *);
-extern wint_t mb_next_wc (char const *, char const *);
 
 /* dfasearch.c */
-extern struct localeinfo localeinfo;
-extern void GEAcompile (char const *, size_t, reg_syntax_t);
-extern size_t EGexecute (char const *, size_t, size_t *, char const *);
+extern void *GEAcompile (char *, size_t, reg_syntax_t);
+extern size_t EGexecute (void *, char const *, size_t, size_t *, char const *);
 
 /* kwsearch.c */
-extern void Fcompile (char const *, size_t);
-extern size_t Fexecute (char const *, size_t, size_t *, char const *);
+extern void *Fcompile (char *, size_t, reg_syntax_t);
+extern size_t Fexecute (void *, char const *, size_t, size_t *, char const *);
 
 /* pcresearch.c */
-extern void Pcompile (char const *, size_t);
-extern size_t Pexecute (char const *, size_t, size_t *, char const *);
+extern void *Pcompile (char *, size_t, reg_syntax_t);
+extern size_t Pexecute (void *, char const *, size_t, size_t *, char const *);
+
+/* grep.c */
+extern struct localeinfo localeinfo;
+extern void fgrep_to_grep_pattern (char **, size_t *);
 
 /* Return the number of bytes in the character at the start of S, which
    is of size N.  N must be positive.  MBS is the conversion state.
